#############################################
# Simulation file for 'Sequential Selection #
# Procedures and FDR Control' (M. Grazier   #
# G'Sell et al.                             #
#############################################
###################
# Mattia Ciollaro #
###################


#############
# functions #
#############
index.sampler <- function( m=100, s=20, probs ){
    # function to sample the indices of the true null
    # hypotheses

    indices <- sample( 1:m, m-s, prob=probs )

    return( indices )

}

probs <- function( a=0.9, m=100, r=0.5, separated=FALSE, s=NULL ){
    # function to generate a vector of probabilities for
    # the sampling without replacement of the indices of
    # the true null hypotheses.
    # Note: highe r means slower geometric decay of the
    # probabilities; a is the largest allowed probability. 

    if( separated == TRUE ){

        probs <- c( rep( 0, s ), rep( 1 / ( m - s ), m-s ) )
        
    }

    else{
    
        exponents <- seq( 0, m-1, by=1 )
        total <- a * ( 1 - r^m ) / ( 1 - r )

        probs <- a * r^exponents / total
        probs <- sort( probs, decreasing=FALSE )
        
}

    return( probs )
    
}

probs.Max <- function( m, gamma=23 ){

    values <- (1:m)^gamma
    probs <- values / sum( values )

    return( probs )

}

thresholding.at.alpha <- function( alpha=0.05, pvalues ){

    if( pvalues[1] > alpha){

        index <- NULL

    }
    
    else{

        index <- which.max( pvalues > alpha ) - 1

        if( index == 0 ){

            index <- NULL

        }
        
    }

    return( index )
    
}

alpha.investing <- function( m=100, alpha=0.05, pvalues ){

        k <- seq( 1, m-1, by = 1 )
        values <- ( k + 1 ) * alpha / ( 1 + ( k + 1 ) * alpha )

        index.test <- pvalues[2:m] > values
        index <- which.max( index.test == TRUE )

        if( index == 1 ){

           test <- ( pvalues[2] > ( 2 * alpha / ( 1 + 2 * alpha ) ) )

           if( test == FALSE ){

               index <- NULL

           }

        }

    return( index )

}

forward.stop <- function( alpha=0.05, pvalues ){

    y <- -log( 1 - pvalues )
    sums <- cumsum( y )
    averages <- sums / ( 1:length( sums ) )

    if( min( averages ) > alpha ){

        index <- NULL
        
    }

    else{
    
        index <- max( which( averages <= alpha ) )

    }

    return( index )

}

strong.stop <- function( alpha=0.05, pvalues ){

    y <- -log( pvalues )
    z <- y / ( 1:length( y ) )
    z <- sum( z ) - cumsum( z )
    ## z <- cumsum( y[length(y):1] / ( length( y ):1 ) )
    ## z <- z[length( z ):1]

    q <- exp( -z )

    values <- alpha * seq( 1, length( pvalues ), by=1 ) / length( pvalues )
                                                             
    if( all( q > values ) ){

        index <- NULL
        
    }

    else{
    
        index <- max( which( q < values ) )

    }

    return( index )

}

one.simulation <- function( m=100, s=20, a=0.9, r=0.5, separated=FALSE, beta=23, probs.function='mine' ){

    # sample the indices of the true null hypotheses
    if( separated == FALSE ){

        if( probs.function=='mine'){
            
            probs <- probs( a, m, r, separated, s )

        }
        if( probs.function=='Max'){

            probs <- probs.Max( m=m, gamma=beta )
            
        }
        true.null.indices <- index.sampler( m, s, probs )
        false.null.indices <- setdiff( 1:m, true.null.indices )

    }

    if( separated == TRUE ){

        true.null.indices <- (s+1):m
        false.null.indices <- 1:s
        
    }

    # generate the p-values
    true.null.pvalues <- runif( m-s )
    false.null.pvalues <- rbeta( s, 1, beta )
    pvalues <- numeric( length=m )
    pvalues[true.null.indices] <- true.null.pvalues
    pvalues[false.null.indices] <- false.null.pvalues    

    return( list( pvalues=pvalues, true.null.indices=sort( true.null.indices ),
                 false.null.indices=sort( false.null.indices ), probs=probs ) )
    
}

whole.simulation <- function( n.simulations=100, m=100, s=20, a=0.9, r=0.5,
                             separated=FALSE, beta=23, alpha=c( 0.05, 0.1, 0.25 ), probs.function='mine'){

    # initialize matrices for simulation results
    power <- matrix( NA, nrow=4, ncol=length( alpha ) )
    fdr <- matrix( NA, nrow=4, ncol=length( alpha ) )
    rownames( power ) <- c( 'alpha.investing', 'alpha.thresholding', 'forward.stop',
                           'strong.stop' )
    colnames( power ) <- as.character( alpha )
    rownames( fdr ) <- rownames( power )
    colnames( fdr ) <- colnames( power )

    pvalues <- matrix( NA, ncol=m, nrow=length( alpha )*n.simulations )
    false.null.indices <- matrix( 0, ncol=m, nrow=length( alpha )*n.simulations )

    # run
    for( i in seq_along( alpha ) ){

        # within simulation initializations
        alpha.investing.power <- numeric( length=n.simulations )
        thresholding.at.alpha.power <- numeric( length=n.simulations )
        forward.stop.power <- numeric( length=n.simulations )
        strong.stop.power <- numeric( length=n.simulations )

        alpha.investing.fdr <- numeric( length=n.simulations )
        thresholding.at.alpha.fdr <- numeric( length=n.simulations )
        forward.stop.fdr <- numeric( length=n.simulations )
        strong.stop.fdr <- numeric( length=n.simulations )

        # run simulation for n.simulations times
        for( j in 1:n.simulations ){

            sim <- one.simulation( m, s, a, r, separated=separated, beta, probs.function=probs.function )
            
            pvalues[(i-1)*n.simulations + j,] <- sim$pvalues
            false.null.indices[(i-1)*n.simulations + j,sim$false.null.indices] <- 1

            # apply rules
            sim.alpha.investing <- alpha.investing( m, alpha[i], sim$pvalues )
            sim.thresholding.at.alpha <- thresholding.at.alpha( alpha[i], sim$pvalues)
            sim.forward.stop <- forward.stop( alpha[i], sim$pvalues )
            sim.strong.stop <-  strong.stop( alpha[i], sim$pvalues )

            # indices with rejected hypotheses
            reject.alpha.investing <- NULL
            if( is.null( sim.alpha.investing ) == FALSE ){

                reject.alpha.investing <- 1:sim.alpha.investing

            }
            reject.thresholding.at.alpha <- NULL
            if( is.null( sim.thresholding.at.alpha ) == FALSE ){

                reject.thresholding.at.alpha <- 1:sim.thresholding.at.alpha

            }
            reject.forward.stop <- NULL
            if( is.null( sim.forward.stop ) == FALSE ){

                reject.forward.stop <- 1:sim.forward.stop

            }
            reject.strong.stop <- NULL
            if( is.null( sim.strong.stop ) == FALSE ){

                reject.strong.stop <- 1:sim.strong.stop

            }

            # power and FDR performance within current simulation
            alpha.investing.power[j] <- 0
            alpha.investing.fdr[j] <- 0
            
            if( is.null( reject.alpha.investing ) == FALSE ){

                alpha.investing.power[j] <- length( intersect( reject.alpha.investing,
                                                              sim$false.null.indices ) ) / s
                alpha.investing.fdr[j] <- length( intersect( reject.alpha.investing,
                                                            sim$true.null.indices ) ) / sim.alpha.investing

            }

            thresholding.at.alpha.power[j] <- 0
            thresholding.at.alpha.fdr[j] <- 0
            if( is.null( reject.thresholding.at.alpha ) == FALSE ){

                thresholding.at.alpha.power[j] <- length( intersect( reject.thresholding.at.alpha,
                                                                    sim$false.null.indices ) ) / s
                thresholding.at.alpha.fdr[j] <- length( intersect( reject.thresholding.at.alpha,
                                                                  sim$true.null.indices ) ) / sim.thresholding.at.alpha

                
            }

            forward.stop.power[j] <- 0
            forward.stop.fdr[j] <- 0
            
            if( is.null( reject.forward.stop ) == FALSE ){

                forward.stop.power[j] <- length( intersect( reject.forward.stop, sim$false.null.indices ) ) / s
                forward.stop.fdr[j] <- length( intersect( reject.forward.stop, sim$true.null.indices ) ) / sim.forward.stop

            }

            strong.stop.power[j] <- 0
            strong.stop.fdr[j] <- 0
            if( is.null( reject.strong.stop ) == FALSE ){

                strong.stop.power[j] <- length( intersect( reject.strong.stop, sim$false.null.indices ) ) / s
                strong.stop.fdr[j] <- length( intersect( reject.strong.stop, sim$true.null.indices ) ) / sim.strong.stop

                
            }
            
        }

        # power and FDR averaged over all simulations
        power[,i] <- c( mean( alpha.investing.power ), mean( thresholding.at.alpha.power ), mean( forward.stop.power ), mean( strong.stop.power ) )
        fdr[,i] <- c( mean( alpha.investing.fdr ), mean( thresholding.at.alpha.fdr ), mean( forward.stop.fdr ), mean( strong.stop.fdr ) )
        
    }

    return( list( alpha=alpha, fdr=fdr, power=power, pvalues=pvalues, false.null.indices=false.null.indices) )
    
}

plot.pvalues <- function( pvalues, false.null.indices, cex=0.8, pch=16, col.true='black', col.false='red', only.first=100, main='' ){

    m <- dim( pvalues )[2]
    K <- min( dim( pvalues )[1], only.first )

    plot( which(false.null.indices[1,]==TRUE), pvalues[1,which(false.null.indices[1,]==TRUE)], ylim=c( 0, 1 ), xlim=c(1,m), cex=cex, pch=pch, col=col.false, xlab='hypothesis index', ylab='p-value', main=main )
    points( which(false.null.indices[1,]==FALSE), pvalues[1,which(false.null.indices[1,]==FALSE)], cex=cex, pch=pch, col=col.true)

    for(i in 2:K){

        points( which(false.null.indices[i,]==TRUE), pvalues[i,which(false.null.indices[i,]==TRUE)], cex=cex, pch=pch, col=col.false )
        points( which(false.null.indices[i,]==FALSE), pvalues[i,which(false.null.indices[i,]==FALSE)], cex=cex, pch=pch, col=col.true )

    }
    
}

plot.simulation.results <- function( simulation, lwd=2, ylim.fdr=c(0,0.3), ylim.power=c(0,1), main='' ){

    alpha <- simulation$alpha
    fdr <- simulation$fdr
    power <- simulation$power
    
    dev.new( )
    par( mfrow=c( 1, 2 ) )

    plot( alpha, alpha, ylim=ylim.fdr, type='l', ylab='FDR', xlab=expression(alpha), main=main )
    lines( alpha, fdr[1,], type='o', col='orange', lwd=lwd )
    lines( alpha, fdr[2,], type='o', col='cyan', lwd=lwd )
    lines( alpha, fdr[3,], type='o', col='green', lwd=lwd )
    lines( alpha, fdr[4,], type='o', col='grey', lwd=lwd )
    legend('topleft',c('alpha investing', 'alpha thresholding', 'ForwardStop', 'StrongStop' ), lwd=lwd, col=c('orange', 'cyan', 'green', 'grey' ) )

    plot( alpha, power[1,], type='o', col='orange', lwd=lwd, ylim=ylim.power, ylab='Power', xlab=expression(alpha), main=main )
    lines( alpha, power[2,], type='o', col='cyan', lwd=lwd )
    lines( alpha, power[3,], type='o', col='green', lwd=lwd )
    lines( alpha, power[4,], type='o', col='grey', lwd=lwd )
    
}